"""Tools for fitting curves to data."""

from collections import OrderedDict
import numpy as np
from matplotlib import pyplot as plt
from scipy.optimize import curve_fit
from scipy import inf

__VERSION__ = [1, 1]

SUPERSCRIPTS = {'-': '⁻', '+': '⁺', '0': '⁰', '1': '¹', '2': '²', '3': '³',
                '4': '⁴', '5': '⁵', '6': '⁶', '7': '⁷', '8': '⁸', '9': '⁹'}


class Fit(object):
    """A class to fit a model to some data.

    Parameters
    ----------
    x_data : np.ndarray
        The values of the independent variable.
    y_data : np.ndarray
        The values of the dependent variable.
    model : func
        A `function` representing a guess at the functional form of the data.
        The function takes the following arguments:
          - A `numpy.ndarray` representing the independent variable.
          - A `float` variable for each fit parameter.
        The function should return a `numpy.ndarray` indicating the values of
        the trial function given the fitting parameters and the independent
        variable.
    initial_parameters : list of float or tuple of float, optional
        A list of floats specifying the initial guesses for the fit
        parameters. If `None` (default) is specified, a list of 1s will be
        created. If an item is a `tuple` of `float` instead of a single
        `float`,
          - The first will be treated as the guess,
          - the second will be treated as a lower bound, and
          - the third will be treated as an upper bound.
    max_runs : int, optional
        The maximum number of iterations to perform to seek convergence.
        The default is 2000.

    Attributes
    ----------
    best_fit_parameters : OrderedDict
        A dictionary containing the results of the fit. The keys of the
        dict are the names of the inputs to `model` (other than the
        independent variable). Each value is a `tuple` of `float`,
        specifying the fit value and the standard error.
    fit_values : dict
        A dictionary containing the best fit values. The keys are the
        names of the inputs to `model` (other than the independent
        variable). The values are the results of the fit.
    fit_errors : dict
        A dictionary containing the errors in the fit. The keys are the
        names of the inputs to `model` (other than the independent
        variable). The values are the errors.
    """

    def __init__(self, x_data, y_data, model, initial_parameters=None,
                 max_runs=2000):
        self.model = model

        if initial_parameters is None:
            self.p0 = None
            self.bounds = (-inf, inf)
        else:
            self.p0 = []
            lower_bounds = []
            upper_bounds = []
            for item in initial_parameters:
                if isinstance(item, (float, int)):
                    self.p0.append(item)
                    lower_bounds.append(-inf)
                    upper_bounds.append(inf)
                else:
                    self.p0.append(item[0])
                    lower_bounds.append(item[1])
                    upper_bounds.append(item[2])
            self.p0 = np.array(self.p0)
            self.bounds = (np.array(lower_bounds), np.array(upper_bounds))

        self.x = x_data
        self.y = y_data
        self.max_runs = max_runs

        self.best_fit_parameters = self._run_fit(self.max_runs)
        self.fit_values = {}
        self.fit_errors = {}
        for key, val in self.best_fit_parameters.items():
            self.fit_values[key] = val[0]
            self.fit_errors[key] = val[1]

    def _run_fit(self, max_runs=None):
        """Execute the fit.

        Parameters
        ----------
        max_runs : int, optional
            The maximum number of iterations to perform in an attempt to
            achieve convergence.

        Returns
        -------
        best_fit_parameters : OrderedDict
            An `OrderedDict` containing the best fit parameters. The
            keys are the names of the inputs to `model`. The values
            are tuples:

            value : float
                The value of the parameter which best fits the data.
            error : float
                The standard error in the respective fit parameter.
        """
        result, pcov = curve_fit(self.model, self.x, self.y, self.p0,
                                 bounds=self.bounds, maxfev=max_runs)

        with np.errstate(invalid='print'):
            errors = np.sqrt(np.diag(pcov))

        parnames = self.model.__code__.co_varnames[1:]

        best_fit_parameters = OrderedDict()
        for name, val, err in zip(parnames, result, errors):
            best_fit_parameters[name] = (val, err)

        return best_fit_parameters

    def calculate_y_data(self, x_data):
        """Return the fitted values of the dependent variable.

        Parameters
        ----------
        x_data : numpy.ndarray
            The values of the independent variable.

        Returns
        -------
        y_data : numpy.ndarray
            The fitted values of the dependent variable.
        """
        params = [v[0] for v in self.best_fit_parameters.values()]
        return self.model(x_data, *params)

    def get_fit_data(self, points=500, overhang=0.02):
        """Return fitted data for plotting.

        Parameters
        ----------
        points : int, optional
            The number of points to calculate (the default is 500).
        overhang : float, optional
            The fraction of the range to extend the data beyond the
            given data (the default is 0.02).

        Returns
        -------
        x : numpy.ndarray
            Data points for the independent variable.
        y : numpy.ndarray
            Data points for the dependent variable according to the
            fit.
        """
        min_x = np.min(self.x)
        if min_x < 0:
            min_x = min_x*(1+overhang)
        else:
            min_x = min_x*(1-overhang)
        max_x = np.max(self.x)
        if max_x > 0:
            max_x = max_x*(1+overhang)
        else:
            max_x = max_x*(1-overhang)
        x = np.linspace(min_x, max_x, points)
        return (x, self.calculate_y_data(x))

    def print_results(self, sig_figs=False):
        """Print the results of the last fit.

        Parameters
        ----------
        sig_figs : bool, optional
            If `True`, print only significant digits. Otherwise,
            print all available digits. The default is `False`.
        """
        if sig_figs:
            self._print_results_sig_figs()
        else:
            self._print_results_all()

    def _print_results_sig_figs(self):
        """Print the results of the fit, showing only significant digits."""
        data = [str(Datum(bfp, err)).split(' ± ') for bfp, err
                in self.best_fit_parameters.values()]
        for i, line in enumerate(data):
            pt1, pt2 = line[1].split(' × ')
            data[i][1] = pt1
            data[i].append(pt2)
        columns = [list(self.best_fit_parameters.keys())]
        columns.append(['=']*len(columns[0]))
        columns.append([d[0] for d in data])
        columns.append(['±']*len(columns[0]))
        columns.append([d[1] for d in data])
        columns.append(['×']*len(columns[0]))
        columns.append([d[2] for d in data])
        percent = []
        for val, err in self.best_fit_parameters.values():
            percent.append(f'({abs((err/val)*100):.3f}%)')
        columns.append(percent)
        t = Table(columns)

        print()
        head = "-- Best Fit Results "
        head = head + ('-'*(t.length()-len(head)))
        print(head)
        t.print()
        print('-'*t.length())

    def _print_results_all(self):
        """Print the results of the last fit, showing all digits."""
        columns = [list(self.best_fit_parameters.keys())]
        columns.append(['=']*len(columns[0]))
        columns.append([])  # index 2, values
        columns.append(['±']*len(columns[0]))
        columns.append([])  # index 4, errors
        columns.append([])  # index 5, percents
        for item in self.best_fit_parameters:
            val, err = self.best_fit_parameters[item]
            columns[2].append(str(val))
            columns[4].append(str(err))
            columns[5].append(f'({abs((err/val)*100):.3f}%)')
        t = Table(columns)

        print()
        head = "-- Best Fit Results "
        head = head + ('-'*(t.length()-len(head)))
        print(head)
        t.print()
        print('-'*t.length())

    def show_plot(self, new_fig=True):
        """Show a plot of the fit.

        Parameters
        ----------
        new_fig : bool
            Whether to create a new figure. If `True` (the default),
            a new figure with the original data and the fit will
            be created. If `False`, the fit data will be added to the
            active axis of `pyplot`.

        Returns
        -------
        fig, ax : tuple of two elements
            The first element is the relevant `mpl.figure.Figure`
            object. The second is the relevant
            `mpl.axes._subplots.AxesSubplot` object.
        """
        x, y = self.get_fit_data()

        if new_fig:
            fig = plt.figure(tight_layout=True)
            ax = fig.add_subplot(111)
            ax.plot(self.x, self.y, '.')
            ax.plot(x, y, '-')
            return (fig, ax)
        else:
            plt.plot(x, y, '-')
            return (plt.gcf(), plt.gca())


def linear(x, m, b):
    """Model for a linear function."""
    return m*x + b


def sine(t, A, ω, ϕ0, y0):
    """Model for a sine function of the form A*sin(ωt + ϕ0) + y0."""
    return A*np.sin(ω*t + ϕ0) + y0


class Datum:
    """A class to represent a single number with uncertainty.

    Parameters
    ----------
    value : float
        The numerical value of the quantity.
    uncertainty : float
        The uncertainty in the value.
    """

    def __init__(self, value, uncertainty):
        if np.isnan(uncertainty):
            self.x = value
            self.dx = uncertainty
            self.power = 0
            return
        power = np.round(np.log10(uncertainty))
        dx = uncertainty/(10**power)
        x = value/(10**power)
        if dx >= 2:
            dx = np.around(dx, 0)
        elif dx >= 1:
            dx = np.around(dx, 1)
            dx *= 10
            x *= 10
            power -= 1
        elif dx > 0.2:
            dx = np.around(dx, 1)
        else:
            dx = np.around(dx, 2)
            dx *= 10
            x *= 10
            power -= 1
        x = value/(10**power)
        if dx < 1:
            dx = dx*10
            x = x*10
            power -= 1
        self.x = int(np.around(x))
        self.dx = int(dx)
        self.power = int(power)

    def get_strings(self):
        """Return strings representing the number and its uncertainty.

        Returns
        -------
        value : float
            The value of the quantity with the correct number of
            significant digits.
        uncertainty : float
            The uncertainty in the quantity.
        """
        try:
            # '%dE%+d' % (self.x, self.power)
            val1 = f'{self.x:d}E{self.power:+d}'
        except ValueError:
            try:
                val1 = str(self.x)
            except ValueError:
                val1 = '????'
        try:
            val2 = f'{self.dx:d}E{self.power:+d}'
        except ValueError:
            val2 = '????'
        return (val1, val2)

    def _normalize(self):
        if self.x == 0:
            adjust = 0
        else:
            adjust = np.floor(np.log10(abs(self.x)))

        output = (self.x/10**adjust, self.dx/10**adjust,
                  int(self.power + adjust))
        return output

    def __str__(self):
        x, dx, power = self._normalize()
        dec_x = str(x).split('.')[-1]
        dec_dx = str(dx).split('.')[-1]
        strpow = f'{power:+d}'
        for key, val in SUPERSCRIPTS.items():
            strpow = strpow.replace(key, val)
        if dec_x == '0' and dec_dx == '0':
            x = int(x)
            dx = int(dx)
            return f'({x} ± {dx}) × 10{strpow}'
        if np.isnan(dx):
            return f'({x:.3f} ± {dx}) × 10{strpow}'
        return f'({x} ± {dx}) × 10{strpow}'


class Table:
    """A class for aligning columns of numbers.

    Parameters
    ----------
    columns : list of list of str
        Each list should represent a column in the table. The
        numbers should have already been converted to strings
        of appropriate length.
    minpad : str
        The string should contain some number of spaces that will
        be the minimum separation between columns.
    """

    def __init__(self, columns, minpad=' '):
        self.widths = []
        for c in columns:
            width = [len(row) for row in c]
            self.widths.append(max(width))

        self.columns = columns
        self.minpad = minpad

    def print(self):
        """Print the table with appropriate spacing."""
        for row in zip(*self.columns):
            for item, width in zip(row, self.widths):
                pad = width + len(self.minpad) - len(item)
                print((' ' * pad) + item, end='')
            print()

    def length(self):
        """Return the full width of the table."""
        return sum(self.widths) + len(self.columns)*len(self.minpad)


if __name__ == '__main__':
    xdat, ydat = np.loadtxt('fittable5.csv', unpack=True, delimiter=',')

    def slit(x, I0, a, d, lamb, L):
        arg1 = np.pi*a*x/(lamb*L)
        arg2 = np.pi*d*x/(lamb*L)
        return I0 * (np.sin(arg1)/arg1)**2 * np.cos(arg2)**2

    f = Fit(xdat, ydat, slit)
    f.print_results(True)
    plt.plot(f.x, f.y)
    f.show_plot()
